import asyncio
from collections import Counter
from typing import Type, Union
from collections.abc import MutableMapping

import pytest

import aiomisc
from aiomisc.circuit_breaker import (
    CircuitBreakerStates as States,
    CircuitBroken,
)

pytestmark = pytest.mark.catch_loop_exceptions


class CallContainer:
    def __init__(self):
        self.failed = False

    def __call__(self, *args, **kwargs):
        if self.failed:
            raise RuntimeError

        return


class PatchedCircuitBreaker(aiomisc.CircuitBreaker):
    """
    CircuitBreaker for young time travelers
    """

    _TIME = 0.0

    @classmethod
    def tick(cls, second=1.0):
        cls._TIME += second

    @classmethod
    def reset(cls):
        cls._TIME += 1000000000.0

    def _get_time(self):
        return float(self._TIME)


async def test_simple(event_loop):
    PatchedCircuitBreaker.reset()
    circuit_breaker = PatchedCircuitBreaker(error_ratio=0.5, response_time=10)

    ctx = CallContainer()

    # Zero ratio on start
    assert circuit_breaker.recovery_ratio == 0

    # PASSING state
    for _ in range(10):
        PatchedCircuitBreaker.tick()
        circuit_breaker.call(ctx)

    assert circuit_breaker.recovery_ratio == 0
    assert circuit_breaker.state == States.PASSING

    # Now context return exceptions
    ctx.failed = True

    # Errored calls
    for _ in range(10):
        PatchedCircuitBreaker.tick()

        with pytest.raises(RuntimeError):
            circuit_breaker.call(ctx)

    # Change state to BROKEN
    with pytest.raises(CircuitBroken):
        circuit_breaker.call(ctx)

    assert circuit_breaker.state == States.BROKEN

    # Delaying in BROKEN state
    assert circuit_breaker.get_state_delay() > 0

    # Waiting in BROKEN state
    for _ in range(10):
        PatchedCircuitBreaker.tick()

        with pytest.raises(CircuitBroken):
            circuit_breaker.call(ctx)

    responses: MutableMapping[bool | type[Exception], int] = Counter()

    # Delay is zero
    assert circuit_breaker.get_state_delay() == 0

    # Collect statistic in RECOVERY state
    for _ in range(110):
        PatchedCircuitBreaker.tick(0.1)

        try:
            circuit_breaker.call(ctx)
        except CircuitBroken:
            responses[CircuitBroken] += 1
        except RuntimeError:
            responses[RuntimeError] += 1

    assert responses[CircuitBroken]
    assert responses[RuntimeError]

    # Waiting to switch from BROKEN to RECOVERY
    PatchedCircuitBreaker.tick(circuit_breaker.get_state_delay())

    ctx.failed = False

    responses.clear()

    # RECOVERY to PASSING state
    for _ in range(10):
        PatchedCircuitBreaker.tick()

        try:
            circuit_breaker.call(ctx)
            responses[True] += 1
        except CircuitBroken:
            responses[CircuitBroken] += 1
        except RuntimeError:
            responses[RuntimeError] += 1

    assert not responses.get(RuntimeError)
    assert responses[CircuitBroken]
    assert responses[True]

    responses.clear()
    ctx.failed = False

    # PASSING state
    for _ in range(20):
        PatchedCircuitBreaker.tick()

        try:
            circuit_breaker.call(ctx)
            responses[True] += 1
        except CircuitBroken:
            responses[CircuitBroken] += 1
        except RuntimeError:
            responses[RuntimeError] += 1

    assert not responses.get(CircuitBroken)
    assert responses[True]


async def test_get_time():
    cb = aiomisc.CircuitBreaker(0.5, 1)
    assert cb._get_time()
    time = cb._get_time()
    await asyncio.sleep(0.02)
    assert cb._get_time() > time


async def test_bad_response_time():
    with pytest.raises(ValueError):
        aiomisc.CircuitBreaker(0.5, 0)


def test_exception_inspector():
    def inspector(exc: Exception) -> bool:
        if isinstance(exc, ValueError):
            return False
        return True

    cb = PatchedCircuitBreaker(
        error_ratio=0.5, response_time=5, exception_inspector=inspector
    )

    PatchedCircuitBreaker.reset()

    def raiser(exc_type):
        raise exc_type()

    for i in range(100):
        PatchedCircuitBreaker.tick(0.1)
        with pytest.raises(ValueError):
            cb.call(raiser, ValueError)

        assert cb.recovery_ratio == 0

    PatchedCircuitBreaker.tick(1)

    for i in range(10):
        PatchedCircuitBreaker.tick(0.1)
        with pytest.raises(RuntimeError):
            cb.call(raiser, RuntimeError)

        assert cb.recovery_ratio > 0



import asyncio
import logging
from time import monotonic

from aiomisc.entrypoint import entrypoint

from .server import RPCServer

log = logging.getLogger("client")


async def main(rpc: RPCServer, server_host: str, server_port: int) -> None:
    call_count = 300

    delta = -monotonic()

    for i in range(call_count):
        await asyncio.gather(
            *[
                asyncio.wait_for(
                    rpc(
                        server_host,
                        server_port,
                        "multiply",
                        x=120000,
                        y=1000000,
                    ),
                    timeout=5,
                )
                for _ in range(call_count)
            ]
        )

    delta += monotonic()

    total_request_sent = call_count**2

    log.info("Total executed %d requests on %.3f", total_request_sent, delta)
    log.info("RPS: %.3f", total_request_sent / delta)

    log.info("Close connection")


if __name__ == "__main__":
    service = RPCServer(address="::", port=0, handlers={})
    with entrypoint(service) as loop:
        loop.run_until_complete(main(service, "::1", 15678))



import asyncio
import os
import ssl
import time
from contextlib import contextmanager
from pathlib import Path

import pytest

import aiomisc

try:
    import uvloop
except ImportError:
    uvloop = None  # type: ignore


@pytest.fixture
def timer():
    @contextmanager
    def timer(expected_time=0, *, dispersion=0.5):
        expected_time = float(expected_time)
        dispersion_value = expected_time * dispersion

        now = time.time()

        yield

        delta = time.time() - now

        lower_bound = expected_time - dispersion_value
        upper_bound = expected_time + dispersion_value

        assert lower_bound < delta < upper_bound

    return timer


def thread_pool_executor(request):
    return aiomisc.ThreadPoolExecutor


policies = (asyncio.DefaultEventLoopPolicy(),)
policy_ids = ("asyncio",)

if uvloop:
    policies = (uvloop.EventLoopPolicy(),) + policies  # type: ignore
    policy_ids = ("uvloop",) + policy_ids  # type: ignore


@pytest.fixture(params=policies, ids=policy_ids)
def event_loop_policy(request):
    return request.param


@pytest.fixture()
def certs():
    return Path(os.path.dirname(os.path.abspath(__file__))) / "certs"


@pytest.fixture
def ssl_client_context(certs):
    ca = str(certs / "ca.pem")
    key = str(certs / "client.key")
    cert = str(certs / "client.pem")

    context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH, capath=ca)

    if key:
        context.load_cert_chain(cert, key)

    context.load_verify_locations(cafile=ca)
    context.check_hostname = False
    context.verify_mode = ssl.VerifyMode.CERT_NONE

    return context



import dnslib  # type: ignore[import-untyped]
import pytest

from aiomisc.service.dns import DNSStore, DNSZone
from aiomisc.service.dns.records import (
    AAAA,
    CAA,
    CNAME,
    DNSKEY,
    DS,
    HTTPS,
    LOC,
    MX,
    NAPTR,
    NS,
    NSEC,
    PTR,
    RP,
    RRSIG,
    SOA,
    SRV,
    SSHFP,
    TLSA,
    TXT,
    A,
    RecordType,
)
from aiomisc.service.dns.tree import RadixTree


def test_insert_and_search():
    tree: RadixTree[str] = RadixTree()
    tree.insert(("com", "example"), "Zone1")
    assert tree.search(("com", "example")) == "Zone1"
    assert tree.search(("com", "example", "sub")) is None


def test_insert_multiple_and_search():
    tree: RadixTree[str] = RadixTree()
    tree.insert(("com", "example"), "Zone1")
    tree.insert(("com", "example", "sub"), "Zone2")
    assert tree.search(("com", "example")) == "Zone1"
    assert tree.search(("com", "example", "sub")) == "Zone2"
    assert tree.search(("com", "example", "sub", "test")) is None


def test_find_prefix_basic():
    tree: RadixTree[str] = RadixTree()
    tree.insert(("com", "example"), "Zone1")
    assert tree.find_prefix(("com", "example")) == (("com", "example"), "Zone1")
    assert tree.find_prefix(("com", "example", "sub")) == (
        ("com", "example"),
        "Zone1",
    )


def test_find_prefix_with_subdomain():
    tree: RadixTree[str] = RadixTree()
    tree.insert(("com", "example"), "Zone1")
    tree.insert(("com", "example", "sub"), "Zone2")
    assert tree.find_prefix(("com", "example", "sub", "test")) == (
        ("com", "example", "sub"),
        "Zone2",
    )
    assert tree.find_prefix(("com", "example", "other")) == (
        ("com", "example"),
        "Zone1",
    )


def test_find_prefix_no_match():
    tree: RadixTree[str] = RadixTree()
    tree.insert(("com", "example"), "Zone1")
    assert tree.find_prefix(("org", "example")) is None
    assert tree.find_prefix(("com", "nonexistent")) is None


def test_insert_and_override():
    tree: RadixTree[str] = RadixTree()
    tree.insert(("com", "example"), "Zone1")
    tree.insert(("com", "example"), "Zone2")
    assert tree.search(("com", "example")) == "Zone2"


def test_deep_insert_and_search():
    tree: RadixTree[str] = RadixTree()
    tree.insert(("com", "example", "sub", "deep", "deeper"), "Zone1")
    assert tree.search(("com", "example", "sub", "deep", "deeper")) == "Zone1"
    assert tree.search(("com", "example", "sub", "deep")) is None


def test_find_prefix_with_partial_matches():
    tree: RadixTree[str] = RadixTree()
    tree.insert(("com", "example"), "Zone1")
    tree.insert(("com", "example", "sub"), "Zone2")
    tree.insert(("com", "example", "sub", "subsub"), "Zone3")
    assert tree.find_prefix(("com", "example", "sub", "subsub", "deep")) == (
        ("com", "example", "sub", "subsub"),
        "Zone3",
    )
    assert tree.find_prefix(("com", "example", "sub", "other")) == (
        ("com", "example", "sub"),
        "Zone2",
    )


def test_edge_cases_empty_key():
    tree: RadixTree[str] = RadixTree()
    tree.insert((), "RootZone")
    assert tree.search(()) == "RootZone"
    assert tree.find_prefix(("any", "key")) == ((), "RootZone")


def test_edge_cases_single_character_keys():
    tree: RadixTree[str] = RadixTree()
    tree.insert(("a",), "ZoneA")
    tree.insert(("b",), "ZoneB")
    assert tree.search(("a",)) == "ZoneA"
    assert tree.search(("b",)) == "ZoneB"
    assert tree.search(("c",)) is None
    assert tree.find_prefix(("a", "key")) == (("a",), "ZoneA")


def test_edge_cases_long_keys():
    tree: RadixTree[str] = RadixTree()
    long_key = tuple(str(i) for i in range(1000))
    tree.insert(long_key, "LongZone")
    assert tree.search(long_key) == "LongZone"
    assert tree.find_prefix(long_key + ("extra",)) == (
        tuple(long_key),
        "LongZone",
    )


@pytest.fixture
def dns_store():
    return DNSStore()


def test_add_zone(dns_store):
    zone = DNSZone("example.com.")
    dns_store.add_zone(zone)
    assert dns_store.get_zone("example.com.") is zone


def test_remove_zone(dns_store):
    zone = DNSZone("example.com.")
    dns_store.add_zone(zone)
    dns_store.remove_zone("example.com.")
    assert dns_store.get_zone("example.com.") is None


def test_add_record_to_zone(dns_store):
    zone = DNSZone("example.com.")
    record = A.create(name="www.example.com.", ip="192.0.2.1")
    zone.add_record(record)
    dns_store.add_zone(zone)
    records = dns_store.query("www.example.com.", RecordType.A)
    assert record in records


def test_query_nonexistent_record(dns_store):
    zone = DNSZone("example.com.")
    record = A.create(name="www.example.com.", ip="192.0.2.1")
    zone.add_record(record)
    dns_store.add_zone(zone)
    records = dns_store.query("nonexistent.example.com.", RecordType.A)
    assert len(records) == 0


def test_add_zone_with_existing_name(dns_store):
    zone = DNSZone("example.com.")
    dns_store.add_zone(zone)
    with pytest.raises(ValueError):
        dns_store.add_zone(zone)


def test_remove_nonexistent_zone(dns_store):
    with pytest.raises(ValueError):
        dns_store.remove_zone("nonexistent.com.")


def test_query_with_subdomain(dns_store):
    zone = DNSZone("example.com.")
    record = A.create(name="sub.example.com.", ip="192.0.2.2")
    zone.add_record(record)
    dns_store.add_zone(zone)
    records = dns_store.query("sub.example.com.", RecordType.A)
    assert record in records


def test_get_zone(dns_store):
    zone = DNSZone("example.com.")
    dns_store.add_zone(zone)
    assert dns_store.get_zone("example.com.") is zone


def test_get_nonexistent_zone(dns_store):
    assert dns_store.get_zone("nonexistent.com.") is None


def test_remove_record_from_zone(dns_store):
    zone = DNSZone("example.com.")
    record = A.create(name="www.example.com.", ip="192.0.2.1")
    zone.add_record(record)
    zone.remove_record(record)
    dns_store.add_zone(zone)
    records = dns_store.query("www.example.com.", RecordType.A)
    assert len(records) == 0


def test_find_prefix(dns_store):
    zone = DNSZone("example.com.")
    dns_store.add_zone(zone)
    assert dns_store.get_zone_for_name("sub.example.com.") == ("com", "example")


def test_a_create():
    record = A.create(name="example.com.", ip="192.0.2.1", ttl=300)
    assert record.name == "example.com."
    assert record.type == RecordType.A
    assert isinstance(record.data, dnslib.A)
    assert record.data.data == tuple(map(int, "192.0.2.1".split(".")))
    assert record.ttl == 300


def test_aaaa_create():
    record = AAAA.create(name="example.com.", ipv6="2001:db8::1", ttl=300)
    assert record.name == "example.com."
    assert record.type == RecordType.AAAA
    assert isinstance(record.data, dnslib.AAAA)
    assert record.data.data == (
        32,
        1,
        13,
        184,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
    )
    assert record.ttl == 300


def test_cname_create():
    record = CNAME.create(
        name="example.com.", label="alias.example.com.", ttl=300
    )
    assert record.name == "example.com."
    assert record.type == RecordType.CNAME
    assert isinstance(record.data, dnslib.CNAME)
    assert record.data.label == "alias.example.com."
    assert record.ttl == 300


def test_mx_create():
    record = MX.create(
        name="example.com.",
        exchange="mail.example.com.",
        preference=10,
        ttl=300,
    )
    assert record.name == "example.com."
    assert record.type == RecordType.MX
    assert isinstance(record.data, dnslib.MX)
    assert record.data.label == "mail.example.com."
    assert record.data.preference == 10
    assert record.ttl == 300


def test_txt_create():
    record = TXT.create(name="example.com.", text="some text", ttl=300)
    assert record.name == "example.com."
    assert record.type == RecordType.TXT
    assert isinstance(record.data, dnslib.TXT)
    assert record.data.data == [b"some text"]
    assert record.ttl == 300


def test_soa_create():
    record = SOA.create(
        name="example.com.",
        mname="ns1.example.com.",
        rname="hostmaster.example.com.",
        serial=1,
        refresh=3600,
        retry=1800,
        expire=1209600,
        minimum=600,
        ttl=300,
    )
    assert record.name == "example.com."
    assert record.type == RecordType.SOA
    assert isinstance(record.data, dnslib.SOA)
    assert record.data.mname == "ns1.example.com."
    assert record.data.rname == "hostmaster.example.com."
    assert record.data.times == (1, 3600, 1800, 1209600, 600)
    assert record.ttl == 300


def test_ns_create():
    record = NS.create(name="example.com.", label="ns1.example.com.", ttl=300)
    assert record.name == "example.com."
    assert record.type == RecordType.NS
    assert isinstance(record.data, dnslib.NS)
    assert record.data.label == "ns1.example.com."
    assert record.ttl == 300


def test_ptr_create():
    record = PTR.create(
        name="1.2.3.4.in-addr.arpa.", label="example.com.", ttl=300
    )
    assert record.name == "1.2.3.4.in-addr.arpa."
    assert record.type == RecordType.PTR
    assert isinstance(record.data, dnslib.PTR)
    assert record.data.label == "example.com."
    assert record.ttl == 300


def test_srv_create():
    record = SRV.create(
        name="example.com.",
        priority=10,
        weight=20,
        port=80,
        target="target.example.com.",
        ttl=300,
    )
    assert record.name == "example.com."
    assert record.type == RecordType.SRV
    assert isinstance(record.data, dnslib.SRV)
    assert record.data.priority == 10
    assert record.data.weight == 20
    assert record.data.port == 80
    assert record.data.target == "target.example.com."
    assert record.ttl == 300


def test_caa_create():
    record = CAA.create(
        name="example.com.",
        flags=0,
        tag="issue",
        value="letsencrypt.org",
        ttl=300,
    )
    assert record.name == "example.com."
    assert record.type == RecordType.CAA
    assert isinstance(record.data, dnslib.CAA)
    assert record.data.flags == 0
    assert record.data.tag == "issue"
    assert record.data.value == "letsencrypt.org"
    assert record.ttl == 300


def test_naptr_create():
    record = NAPTR.create(
        name="example.com.",
        order=100,
        preference=10,
        flags="S",
        service="SIP+D2U",
        regexp="",
        replacement=".",
        ttl=300,
    )
    assert record.name == "example.com."
    assert record.type == RecordType.NAPTR
    assert isinstance(record.data, dnslib.NAPTR)
    assert record.data.order == 100
    assert record.data.preference == 10
    assert record.data.flags == "S"
    assert record.data.service == "SIP+D2U"
    assert record.data.regexp == ""
    assert record.data.replacement == "."
    assert record.ttl == 300


def test_ds_create():
    record = DS.create(
        name="example.com.",
        key_tag=12345,
        algorithm=5,
        digest_type=1,
        digest="abc123",
        ttl=300,
    )
    assert record.name == "example.com."
    assert record.type == RecordType.DS
    assert isinstance(record.data, dnslib.DS)
    assert record.data.key_tag == 12345
    assert record.data.algorithm == 5
    assert record.data.digest_type == 1
    assert record.data.digest == b"abc123"
    assert record.ttl == 300


def test_dnskey_create():
    record = DNSKEY.create(
        name="example.com.",
        flags=256,
        protocol=3,
        algorithm=5,
        key="abcdefg",
        ttl=300,
    )
    assert record.name == "example.com."
    assert record.type == RecordType.DNSKEY
    assert isinstance(record.data, dnslib.DNSKEY)
    assert record.data.flags == 256
    assert record.data.protocol == 3
    assert record.data.algorithm == 5
    assert record.data.key == b"abcdefg"
    assert record.ttl == 300


def test_rrsig_create():
    record = RRSIG.create(
        name="example.com.",
        type_covered=RecordType.A,
        algorithm=5,
        labels=2,
        original_ttl=300,
        expiration=1234567890,
        inception=1234560000,
        key_tag=12345,
        signer="example.com.",
        signature="abcdefg",
        ttl=300,
    )
    assert record.name == "example.com."
    assert record.type == RecordType.RRSIG
    assert isinstance(record.data, dnslib.RRSIG)
    assert record.data.covered == RecordType.A
    assert record.data.algorithm == 5
    assert record.data.labels == 2
    assert record.data.orig_ttl == 300
    assert record.data.sig_exp == 1234567890
    assert record.data.sig_inc == 1234560000
    assert record.data.key_tag == 12345
    assert record.data.name == "example.com."
    assert record.data.sig == "abcdefg"
    assert record.ttl == 300


def test_nsec_create():
    record = NSEC.create(
        name="example.com.",
        next_domain="next.example.com.",
        rrtypes=[RecordType.A, RecordType.AAAA],
        ttl=300,
    )
    assert record.name == "example.com."
    assert record.type == RecordType.NSEC
    assert isinstance(record.data, dnslib.NSEC)
    assert record.data.label == "next.example.com."
    assert record.data.rrlist == [RecordType.A, RecordType.AAAA]
    assert record.ttl == 300


def test_https_create():
    record = HTTPS.create(
        name="example.com.",
        priority=10,
        target="target.example.com.",
        params=["param1", "param2"],
        ttl=300,
    )
    assert record.name == "example.com."
    assert record.type == RecordType.HTTPS
    assert isinstance(record.data, dnslib.HTTPS)
    assert record.data.priority == 10
    assert record.data.target == "target.example.com."
    assert record.data.params == ["param1", "param2"]
    assert record.ttl == 300


def test_loc_create():
    record = LOC.create(
        name="example.com.",
        latitude=12.34,
        longitude=56.78,
        altitude=90.0,
        size=1.0,
        h_precision=2.0,
        v_precision=3.0,
        ttl=300,
    )
    assert record.name == "example.com."
    assert record.type == RecordType.LOC
    assert isinstance(record.data, dnslib.LOC)
    assert record.data.lat == "12 20 24.000 N"
    assert record.data.lon == "56 46 48.000 E"
    assert record.data.alt == 90.0
    assert record.data.siz == 1.0
    assert record.data.hp == 2.0
    assert record.data.vp == 3.0
    assert record.ttl == 300


def test_rp_create():
    record = RP.create(
        name="example.com.",
        mbox="hostmaster.example.com.",
        txt="contact.example.com.",
        ttl=300,
    )
    assert record.name == "example.com."
    assert record.type == RecordType.RP
    assert isinstance(record.data, dnslib.RP)
    assert record.data.mbox == "hostmaster.example.com."
    assert record.data.txt == "contact.example.com."
    assert record.ttl == 300


def test_tlsa_create():
    record = TLSA.create(
        name="example.com.",
        usage=1,
        selector=1,
        mtype=1,
        cert="abcdefg",
        ttl=300,
    )
    assert record.name == "example.com."
    assert record.type == RecordType.TLSA
    assert isinstance(record.data, dnslib.TLSA)
    assert record.data.cert_usage == 1
    assert record.data.selector == 1
    assert record.data.matching_type == 1
    assert record.data.cert_data == b"abcdefg"
    assert record.ttl == 300


def test_sshfp_create():
    record = SSHFP.create(
        name="example.com.",
        algorithm=1,
        fptype=1,
        fingerprint="abcdefg",
        ttl=300,
    )
    assert record.name == "example.com."
    assert record.type == RecordType.SSHFP
    assert isinstance(record.data, dnslib.SSHFP)
    assert record.data.algorithm == 1
    assert record.data.fp_type == 1
    assert record.data.fingerprint == b"abcdefg"
    assert record.ttl == 300


def test_zone_replace(dns_store):
    zone = DNSZone("example.com.")
    record1 = A.create(name="www.example.com.", ip="192.0.2.1")
    record2 = A.create(name="api.example.com.", ip="192.0.2.2")
    zone.add_record(record1)
    dns_store.add_zone(zone)

    zone.replace([record2])

    records = dns_store.query("www.example.com.", RecordType.A)
    assert len(records) == 0
    records = dns_store.query("api.example.com.", RecordType.A)
    assert len(records) == 1
    assert record2 in records


def test_zone_replace_multiple_records():
    zone = DNSZone("example.com.")
    record1 = A.create(name="www.example.com.", ip="192.0.2.1")
    record2 = A.create(name="www.example.com.", ip="192.0.2.2")

    zone.replace([record1, record2])
    records = zone.get_records("www.example.com.", RecordType.A)
    assert len(records) == 2
    assert record1 in records
    assert record2 in records


def test_zone_replace_empty():
    zone = DNSZone("example.com.")
    record = A.create(name="www.example.com.", ip="192.0.2.1")
    zone.add_record(record)

    zone.replace([])
    records = zone.get_records("www.example.com.", RecordType.A)
    assert len(records) == 0


def test_zone_replace_invalid_record():
    zone = DNSZone("example.com.")
    record = A.create(name="www.other.com.", ip="192.0.2.1")

    with pytest.raises(ValueError, match="does not belong to zone"):
        zone.replace([record])


def test_store_replace_basic(dns_store):
    zone1 = DNSZone("example.com.")
    record1 = A.create(name="www.example.com.", ip="192.0.2.1")
    zone1.add_record(record1)
    dns_store.add_zone(zone1)

    zone2 = DNSZone("test.com.")
    record2 = A.create(name="www.test.com.", ip="192.0.2.2")
    zone2.add_record(record2)
    dns_store.add_zone(zone2)

    # Replace with new data
    new_record1 = A.create(name="api.example.com.", ip="192.0.2.3")
    new_record2 = A.create(name="api.test.com.", ip="192.0.2.4")

    dns_store.replace(
        {"example.com.": [new_record1], "test.com.": [new_record2]}
    )

    # Check old records are gone
    records = dns_store.query("www.example.com.", RecordType.A)
    assert len(records) == 0
    records = dns_store.query("www.test.com.", RecordType.A)
    assert len(records) == 0

    # Check new records are present
    records = dns_store.query("api.example.com.", RecordType.A)
    assert len(records) == 1
    assert new_record1 in records
    records = dns_store.query("api.test.com.", RecordType.A)
    assert len(records) == 1
    assert new_record2 in records


def test_store_replace_empty(dns_store):
    zone = DNSZone("example.com.")
    record = A.create(name="www.example.com.", ip="192.0.2.1")
    zone.add_record(record)
    dns_store.add_zone(zone)

    dns_store.replace({})

    assert dns_store.get_zone("example.com.") is None
    records = dns_store.query("www.example.com.", RecordType.A)
    assert len(records) == 0


def test_store_replace_multiple_records_per_zone(dns_store):
    new_record1 = A.create(name="www.example.com.", ip="192.0.2.1")
    new_record2 = A.create(name="www.example.com.", ip="192.0.2.2")

    dns_store.replace({"example.com.": [new_record1, new_record2]})

    records = dns_store.query("www.example.com.", RecordType.A)
    assert len(records) == 2
    assert new_record1 in records
    assert new_record2 in records



import asyncio
import os
import sys
from pstats import Stats
from tempfile import NamedTemporaryFile

import pytest

from aiomisc.service.profiler import Profiler
from tests import unix_only


async def test_profiler_start_stop():
    profiler = Profiler(interval=0.1, top_results=10)
    try:
        await profiler.start()
        await asyncio.sleep(0.5)
    finally:
        await profiler.stop()


@unix_only
@pytest.mark.skipif(
    sys.version_info < (3, 10), reason="get_stats_profile available on 3.10+"
)
async def test_profiler_dump():
    profiler = None
    fl = NamedTemporaryFile(delete=False)
    path = NamedTemporaryFile(delete=False).name
    fl.close()
    try:
        profiler = Profiler(interval=0.1, top_results=10, path=path)
        await profiler.start()

        # Get first update
        await asyncio.sleep(0.01)
        stats1 = Stats(path)

        # Not enough sleep till next update
        await asyncio.sleep(0.01)
        stats2 = Stats(path)

        # Getting the same dump
        assert stats1.get_stats_profile() == stats2.get_stats_profile()

        # Enough sleep till next update
        await asyncio.sleep(0.2)
        stats3 = Stats(path)

        # Getting updated dump
        assert stats2.get_stats_profile() != stats3.get_stats_profile()

    finally:
        if profiler:
            await profiler.stop()
        os.remove(path)



import asyncio
import logging
from time import monotonic
from types import MappingProxyType

from aiomisc.entrypoint import entrypoint
from aiomisc.service import RobustTCPClient

from .spec import RPCBase

log = logging.getLogger("client")


class RPCClient(RobustTCPClient, RPCBase):
    async def handle_connection(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        await self.communicate(reader, writer)

    async def start(self) -> None:
        await RPCBase.start(self)
        await RobustTCPClient.start(self)


async def main(client: RPCClient):
    call_count = 1000

    delta = -monotonic()

    for i in range(call_count):
        await asyncio.gather(
            *[
                client("multiply", x=120000, y=1000000)
                for _ in range(call_count)
            ]
        )

    delta += monotonic()

    total_request_sent = call_count**2

    log.info("Total executed %d requests on %.3f", total_request_sent, delta)
    log.info("RPS: %.3f", total_request_sent / delta)
    log.info("Close connection")


handlers = MappingProxyType({"print": print})


if __name__ == "__main__":
    client = RPCClient("::1", 5678, handlers=handlers)
    with entrypoint(client) as loop:
        loop.run_until_complete(main(client))



import asyncio
import json
import logging
import socket
import time
import uuid
from random import shuffle
from typing import List, Optional
from collections.abc import Sequence

import pytest

import aiomisc

pytestmark = pytest.mark.catch_loop_exceptions


async def test_select(event_loop: asyncio.AbstractEventLoop):
    f_one = asyncio.Event()
    f_two = asyncio.Event()

    event_loop.call_soon(f_one.set)
    event_loop.call_later(1, f_two.set)

    two, one = await aiomisc.select(f_two.wait(), f_one.wait())

    assert one
    assert two is None

    one, two = await aiomisc.select(f_one.wait(), f_two.wait())
    assert one


async def test_select_cancelling(event_loop: asyncio.AbstractEventLoop):
    results: list[bool | None] = []

    async def good_coro(wait):
        try:
            if wait:
                await asyncio.sleep(10)
            else:
                await asyncio.sleep(0)
                results.append(True)
                return True
        finally:
            results.append(None)

    one, two = await aiomisc.select(good_coro(False), good_coro(True))
    assert one
    assert results[0]
    assert results[1] is None


async def test_select_exception(event_loop: asyncio.AbstractEventLoop):
    event = asyncio.Event()

    async def bad_coro():
        if event.is_set():
            await asyncio.sleep(10)
        else:
            await asyncio.sleep(0)
            raise ZeroDivisionError

    with pytest.raises(ZeroDivisionError):
        await aiomisc.select(bad_coro(), bad_coro())


@aiomisc.timeout(1)
async def test_select_cancel_false(event_loop: asyncio.AbstractEventLoop):
    event1 = asyncio.Event()
    event2 = asyncio.Event()
    event3 = asyncio.Event()

    async def coro1():
        await event1.wait()
        event2.set()

    async def coro2():
        event1.set()
        await event2.wait()
        event3.set()

    await aiomisc.select(coro2(), coro1(), cancel=False)
    await event3.wait()


@aiomisc.timeout(1)
async def test_select_cancel_true(event_loop: asyncio.AbstractEventLoop):
    event1 = asyncio.Event()
    event2 = asyncio.Event()
    event3 = asyncio.Event()

    async def coro1():
        await event1.wait()
        event2.set()

    async def coro2():
        await asyncio.sleep(0)
        event3.set()

    await aiomisc.select(coro2(), coro1(), cancel=True)
    assert not event1.is_set()
    assert not event2.is_set()


def test_shield():
    results = []

    @aiomisc.shield
    async def coro():
        await asyncio.sleep(0.5)
        results.append(True)

    async def main(event_loop):
        task = event_loop.create_task(coro())
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass
        finally:
            await asyncio.sleep(1)

    with aiomisc.entrypoint() as event_loop:
        event_loop.run_until_complete(
            asyncio.wait_for(main(event_loop), timeout=10)
        )

    assert results == [True]


def test_chunk_list():
    data: Sequence[Sequence[int]] = tuple(
        map(tuple, aiomisc.chunk_list(range(10), 3))
    )

    assert data == ((0, 1, 2), (3, 4, 5), (6, 7, 8), (9,))


def test_configure_logging_json(capsys):
    data = str(uuid.uuid4())

    aiomisc.log.basic_config(
        level=logging.DEBUG, log_format="json", buffered=False
    )
    logging.info(data)

    time.sleep(0.3)
    stdout, stderr = capsys.readouterr()

    json_result = json.loads(stdout.strip())
    assert json_result["msg"] == data

    logging.basicConfig(handlers=[], level=logging.INFO)


def test_configure_logging_stderr(capsys):
    data = str(uuid.uuid4())

    out, err = capsys.readouterr()

    # logging.basicConfig(level=logging.INFO)
    aiomisc.log.basic_config(
        level=logging.DEBUG, log_format="stream", buffered=False
    )

    logging.info(data)

    time.sleep(0.3)
    stdout, stderr = capsys.readouterr()

    assert data in stderr

    logging.basicConfig(handlers=[])


BIND_CASES = [
    ("127.0.0.1", socket.AF_INET),
    ("0.0.0.0", socket.AF_INET),
    ("::", socket.AF_INET6),
]


@pytest.mark.parametrize("address,family", BIND_CASES)
def test_bind_address(address, family, aiomisc_unused_port):
    sock = aiomisc.bind_socket(address=address, port=aiomisc_unused_port)

    assert isinstance(sock, socket.socket)
    assert sock.family == family


async def test_cancel_tasks(event_loop):
    semaphore = asyncio.Semaphore(10)

    tasks = [event_loop.create_task(semaphore.acquire()) for _ in range(20)]

    done, pending = await asyncio.wait(tasks, timeout=0.5)

    for task in pending:
        assert not task.done()

    for task in done:
        assert task.done()

    await aiomisc.cancel_tasks(pending)

    assert len(done) == 10
    assert len(pending) == 10

    for task in pending:
        assert task.done()

        with pytest.raises(asyncio.CancelledError):
            await task


async def test_cancel_tasks_futures(event_loop):
    counter = 0

    def create_future():
        nonlocal counter

        future = event_loop.create_future()
        counter += 1

        if counter <= 10:
            event_loop.call_soon(future.set_result, True)

        return future

    tasks = [create_future() for _ in range(20)]

    await asyncio.sleep(0.5)

    done = tasks[:10]
    pending = tasks[10:]

    for task in pending:
        assert not task.done()

    for task in done:
        assert task.done()

    await aiomisc.cancel_tasks(pending)

    assert len(done) == 10
    assert len(pending) == 10

    for task in pending:
        assert type(task) is asyncio.Future
        assert task.done()

        with pytest.raises(asyncio.CancelledError):
            await task


async def test_cancel_tasks_futures_and_tasks(event_loop):
    tasks = []

    counter = 0

    def create_future():
        nonlocal counter

        future = event_loop.create_future()
        counter += 1

        if counter <= 10:
            event_loop.call_soon(future.set_result, True)

        return future

    for _ in range(20):
        tasks.append(create_future())

    semaphore = asyncio.Semaphore(10)

    for _ in range(20):
        tasks.append(event_loop.create_task(semaphore.acquire()))

    shuffle(tasks)

    done, pending = await asyncio.wait(tasks, timeout=0.5)

    for task in pending:
        assert not task.done()

    for task in done:
        assert task.done()

    await aiomisc.cancel_tasks(pending)

    assert len(done) == 20
    assert len(pending) == 20

    for task in pending:
        assert task.done()

        with pytest.raises(asyncio.CancelledError):
            await task


async def test_awaitable_decorator(event_loop):
    future = event_loop.create_future()
    event_loop.call_soon(future.set_result, 654321)

    @aiomisc.awaitable
    def no_awaitable():
        return 654321

    @aiomisc.awaitable
    def pass_future():
        return future

    @aiomisc.awaitable
    async def coro():
        return await future

    assert pass_future() is future

    assert (await coro()) == 654321
    assert (await pass_future()) == 654321
    assert (await no_awaitable()) == 654321


def test_create_default_event_loop():
    loop, _ = aiomisc.utils.create_default_event_loop()

    async def run():
        with pytest.raises(RuntimeError):
            aiomisc.utils.create_default_event_loop()

    loop.run_until_complete(run())



import asyncio
import logging.handlers
import threading
import traceback
import warnings
from collections.abc import Callable, Iterable
from contextlib import suppress
from queue import Empty, Queue
from socket import socket
from typing import Any, Union

import aiomisc_log
from aiomisc_log.enum import LogFormat, LogLevel

from .counters import Statistic


class ThreadedHandlerStatistic(Statistic):
    threads: int
    records: int
    errors: int
    flushes: int


class ThreadedHandler(logging.Handler):
    def __init__(
        self,
        target: logging.Handler,
        flush_interval: float = 0.1,
        buffered: bool = True,
        queue_size: int = 0,
    ):
        super().__init__()
        self._buffered = buffered
        self._target = target
        self._flush_interval = flush_interval
        self._flush_event = threading.Event()
        self._queue: Queue[logging.LogRecord | None] = Queue(queue_size)
        self._close_event = threading.Event()
        self._thread = threading.Thread(target=self._in_thread, daemon=True)
        self._statistic = ThreadedHandlerStatistic()

    def start(self) -> None:
        self._statistic.threads += 1
        self._thread.start()

    def close(self) -> None:
        self._queue.put(None)
        del self._queue
        self.flush()
        self._close_event.set()
        super().close()

    def flush(self) -> None:
        self._statistic.flushes += 1
        self._flush_event.set()

    def emit(self, record: logging.LogRecord) -> None:
        if self._buffered:
            self._queue.put_nowait(record)
        else:
            self._queue.put(record)
        self._statistic.records += 1

    def _in_thread(self) -> None:
        queue = self._queue
        while not self._close_event.is_set():
            self._flush_event.wait(self._flush_interval)
            try:
                self.acquire()
                while True:
                    record = queue.get(timeout=self._flush_interval)
                    if record is None:
                        return
                    with suppress(Exception):
                        self._target.handle(record)
            except Empty:
                pass
            finally:
                self.release()
        self._statistic.threads -= 1


def suppressor(
    callback: Callable[..., None],
    exceptions: tuple[type[BaseException], ...] = (Exception,),
) -> Callable[..., None]:
    def wrapper() -> None:
        with suppress(*exceptions):
            callback()

    return wrapper


def wrap_logging_handler(
    handler: logging.Handler,
    buffer_size: int = 1024,
    flush_interval: float | int = 0.1,
    loop: asyncio.AbstractEventLoop | None = None,
) -> logging.Handler:
    warnings.warn("wrap_logging_handler is deprecated", DeprecationWarning)
    handler = ThreadedHandler(
        target=handler, queue_size=buffer_size, flush_interval=flush_interval
    )
    handler.start()
    return handler


class UnhandledLoopHook(aiomisc_log.UnhandledHook):
    @staticmethod
    def _fill_transport_extra(
        transport: asyncio.Transport | None, extra: dict[str, Any]
    ) -> None:
        if transport is None:
            return

        extra["transport"] = repr(transport)

        for key in (
            "peername",
            "sockname",
            "compression",
            "cipher",
            "peercert",
            "pipe",
            "subprocess",
        ):
            value = transport.get_extra_info(key)
            if value:
                extra[f"transport_{key}"] = value

    def __call__(
        self, loop: asyncio.AbstractEventLoop, context: dict[str, Any]
    ) -> None:
        context = dict(context)
        message: str = context.pop("message", "unhandled loop exception")
        exception: BaseException | None = context.pop("exception", None)
        future: asyncio.Future | None = context.pop("future", None)
        task: asyncio.Task | None = context.pop("task", None)
        handle: asyncio.Handle | None = context.pop("handle", None)
        protocol: asyncio.Protocol | None = context.pop("protocol", None)
        transport: asyncio.Transport | None = context.pop("transport", None)
        sock: socket | None = context.pop("socket", None)
        source_tb: list[traceback.FrameSummary] = (
            context.pop("source_traceback", None) or []
        )

        if exception is None:
            if future is not None:
                exception = future.exception()
            elif task is not None and task.done():
                exception = task.exception()

        extra = context
        if handle is not None:
            extra["handle"] = repr(handle)
        if protocol is not None:
            extra["protocol"] = repr(protocol)
        if sock is not None:
            extra["sock"] = repr(sock)

        self._fill_transport_extra(transport, extra)
        self.logger.exception(message, exc_info=exception, extra=extra)
        if source_tb:
            self.logger.error("".join(traceback.format_list(source_tb)))


def basic_config(
    level: int | str = LogLevel.default(),
    log_format: str | LogFormat = LogFormat.default(),
    buffered: bool = True,
    buffer_size: int = 0,
    flush_interval: int | float = 0.2,
    loop: asyncio.AbstractEventLoop | None = None,
    handlers: Iterable[logging.Handler] = (),
    **kwargs: Any,
) -> None:
    unhandled_hook = UnhandledLoopHook(logger_name="asyncio.unhandled")

    if loop is None:
        loop = asyncio.get_event_loop()

    forever_task = asyncio.gather(loop.create_future(), return_exceptions=True)
    loop.set_exception_handler(unhandled_hook)

    log_handlers = []

    for user_handler in handlers:
        handler = ThreadedHandler(
            buffered=buffered,
            flush_interval=flush_interval,
            queue_size=buffer_size,
            target=user_handler,
        )
        unhandled_hook.add_handler(handler)
        forever_task.add_done_callback(lambda _: handler.close())
        log_handlers.append(handler)
        handler.start()

    aiomisc_log.basic_config(
        level=level, log_format=log_format, handlers=log_handlers, **kwargs
    )


__all__ = ("LogFormat", "LogLevel", "ThreadedHandler", "basic_config")



import asyncio
import logging
import operator
import sys
import threading
from multiprocessing.context import ProcessError
from os import getpid
from time import sleep
from collections.abc import AsyncGenerator

import pytest
from setproctitle import setproctitle

import aiomisc
from aiomisc import WorkerPool
from aiomisc.worker_pool import WorkerPoolStatistic

PROCESS_NUM = 4


@pytest.fixture
async def worker_pool(
    request: pytest.FixtureRequest, event_loop: asyncio.AbstractEventLoop
) -> AsyncGenerator[WorkerPool, None]:
    async with WorkerPool(
        PROCESS_NUM,
        initializer=setproctitle,
        initializer_args=(f"[WorkerPool] {request.node.name}",),
        statistic_name=request.node.name,
    ) as pool:
        yield pool


@aiomisc.timeout(5)
async def test_success(worker_pool):
    results = await asyncio.gather(
        *[
            worker_pool.create_task(operator.mul, i, i)
            for i in range(worker_pool.workers * 2)
        ]
    )

    results = sorted(results)

    assert sorted(results) == [i * i for i in range(worker_pool.workers * 2)]


@aiomisc.timeout(5)
async def test_incomplete_task_kill(worker_pool):
    await asyncio.gather(
        *[
            worker_pool.create_task(getpid)
            for _ in range(worker_pool.workers * 4)
        ]
    )

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(
            asyncio.gather(
                *[
                    worker_pool.create_task(sleep, 600)
                    for _ in range(worker_pool.workers)
                ]
            ),
            timeout=1,
        )

    await asyncio.gather(
        *[
            worker_pool.create_task(getpid)
            for _ in range(worker_pool.workers * 4)
        ]
    )


@pytest.mark.skipif(
    sys.platform in ("win32", "cygwin"), reason="Windows is ill"
)
@aiomisc.timeout(5)
async def test_incomplete_task_pool_reuse(request, worker_pool: WorkerPool):
    def get_stats() -> dict:
        stats = {}
        for metric in aiomisc.get_statistics(WorkerPoolStatistic):
            if metric.name == request.node.name:
                stats[metric.metric] = metric.value
        return stats

    stats = get_stats()
    while stats["spawning"] < PROCESS_NUM:
        await asyncio.sleep(0.1)
        stats = get_stats()

    pids_start = set(worker_pool.pids)

    await asyncio.gather(
        *[
            worker_pool.create_task(getpid)
            for _ in range(worker_pool.workers * 4)
        ]
    )

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(
            asyncio.gather(
                *[
                    worker_pool.create_task(sleep, 600)
                    for _ in range(worker_pool.workers)
                ]
            ),
            timeout=1,
        )

    await asyncio.gather(
        *[
            worker_pool.create_task(getpid)
            for _ in range(worker_pool.workers * 4)
        ]
    )

    pids_end = set(worker_pool.pids)

    assert list(pids_start) == list(pids_end)


@aiomisc.timeout(5)
async def test_exceptions(worker_pool):
    results = await asyncio.gather(
        *[
            worker_pool.create_task(operator.truediv, i, 0)
            for i in range(worker_pool.workers * 2)
        ],
        return_exceptions=True,
    )

    assert len(results) == worker_pool.workers * 2

    for exc in results:
        assert isinstance(exc, ZeroDivisionError)


@aiomisc.timeout(5)
async def test_exit(worker_pool):
    exceptions = await asyncio.wait_for(
        asyncio.gather(
            *[
                worker_pool.create_task(exit, 1)
                for _ in range(worker_pool.workers)
            ],
            return_exceptions=True,
        ),
        timeout=5,
    )

    assert len(exceptions) == worker_pool.workers
    for exc in exceptions:
        assert isinstance(exc, ProcessError)


@aiomisc.timeout(30)
async def test_exit_respawn(worker_pool):
    exceptions = await asyncio.gather(
        *[
            worker_pool.create_task(exit, 1)
            for _ in range(worker_pool.workers * 3)
        ],
        return_exceptions=True,
    )
    assert len(exceptions) == worker_pool.workers * 3

    for exc in exceptions:
        assert isinstance(exc, ProcessError)


INITIALIZER_ARGS = None
INITIALIZER_KWARGS = None


def initializer(*args, **kwargs):
    global INITIALIZER_ARGS, INITIALIZER_KWARGS
    INITIALIZER_ARGS = args
    INITIALIZER_KWARGS = kwargs
    logging.info("Initializer done")


def get_initializer_args():
    return INITIALIZER_ARGS, INITIALIZER_KWARGS


@aiomisc.timeout(10)
async def test_initializer(worker_pool):
    pool = WorkerPool(
        1,
        initializer=initializer,
        initializer_args=("foo",),
        initializer_kwargs={"spam": "egg"},
    )
    async with pool:
        args, kwargs = await pool.create_task(get_initializer_args)

    assert args == ("foo",)
    assert kwargs == {"spam": "egg"}

    async with WorkerPool(1, initializer=initializer) as pool:
        args, kwargs = await pool.create_task(get_initializer_args)

    assert args == ()
    assert kwargs == {}

    async with WorkerPool(1) as pool:
        args, kwargs = await pool.create_task(get_initializer_args)

    assert args is None
    assert kwargs is None


def bad_initializer():
    return 1 / 0


@aiomisc.timeout(5)
async def test_bad_initializer(worker_pool):
    pool = WorkerPool(1, initializer=bad_initializer)

    with pytest.raises(ZeroDivisionError):
        async with pool:
            await pool.create_task(get_initializer_args)


@aiomisc.timeout(5)
async def test_threads_active_count_in_pool(worker_pool):
    threads = await worker_pool.create_task(threading.active_count)
    assert threads == 1



from typing import Any

import pytest
from aiohttp import web
from yarl import URL

import aiomisc
from aiomisc import bind_socket
from aiomisc.service.aiohttp import AIOHTTPService
from aiomisc.service.raven import RavenSender

pytestmark = pytest.mark.catch_loop_exceptions


class REST(AIOHTTPService):
    @staticmethod
    async def handle(request: web.Request) -> Any:
        raise ValueError("Some error in handle handler")

    async def create_application(self):
        app = web.Application()
        app.router.add_route("GET", "/exc", self.handle)
        app.router.add_route("GET", "/project-test", self.handle)
        return app


@pytest.fixture
def mock_sentry(localhost, aiomisc_unused_port_factory):
    port = aiomisc_unused_port_factory()
    with bind_socket(address=localhost, port=port) as sock:
        rest = REST(sock=sock)
        yield (
            rest,
            URL.build(
                scheme="http",
                user="test",
                password="test",
                port=port,
                host=localhost,
                path="/project-test",
            ),
        )


def test_raven_create(mock_sentry):
    mock_sentry_service, mock_sentry_url = mock_sentry
    svc = RavenSender(sentry_dsn=mock_sentry_url)

    with aiomisc.entrypoint(svc, mock_sentry_service):
        pass



import asyncio

import pytest

import aiomisc
from aiomisc import Service, Signal, entrypoint, receiver


@pytest.fixture
def signal():
    return Signal()


@pytest.fixture(autouse=True, scope="module")
def clear_entrypoint_signals():
    entrypoint.PRE_START = Signal()
    entrypoint.POST_STOP = Signal()


class FooService(Service):
    async def start(self): ...


def test_pre_start_signal(event_loop):
    expected_services = frozenset(FooService() for _ in range(2))
    ep = entrypoint(*expected_services, loop=event_loop)
    received_services = None

    async def pre_start_callback(services, entrypoint):
        nonlocal received_services
        received_services = frozenset(services)

    ep.pre_start.connect(pre_start_callback)

    with ep:
        assert received_services == expected_services


def test_post_stop_signal(event_loop):
    ep = entrypoint(FooService(), loop=event_loop)
    called = False

    async def post_stop_callback(*_, **__):
        nonlocal called
        await asyncio.sleep(0.01)
        called = True

    ep.post_stop.connect(post_stop_callback)

    with ep:
        assert not called

    assert called


def test_entrypoint_class_pre_start_signal(event_loop):
    received_services, received_entrypoint = None, None

    async def pre_start_callback(*, services, entrypoint):
        nonlocal received_services, received_entrypoint
        received_services = services
        received_entrypoint = entrypoint

    entrypoint.PRE_START.connect(pre_start_callback)

    expected_services = (FooService(),)
    ep = entrypoint(*expected_services, loop=event_loop)
    with ep:
        assert received_services == expected_services
        assert received_entrypoint == ep


def test_entrypoint_class_post_stop_signal(event_loop):
    received_entrypoint = None

    async def post_stop_callback(*, services, entrypoint):
        nonlocal received_entrypoint
        received_entrypoint = entrypoint

    entrypoint.POST_STOP.connect(post_stop_callback)

    ep = entrypoint(FooService(), loop=event_loop)

    with ep:
        ...

    assert received_entrypoint == ep


def test_connect_to_frozen_signal(signal):
    signal.freeze()

    async def cb(): ...

    with pytest.raises(RuntimeError):
        signal.connect(cb)


@pytest.mark.parametrize("callback", [None, max, lambda x: x])
def test_wrong_callback(signal, callback):
    with pytest.raises(RuntimeError):
        signal.connect(callback)


async def test_receiver_decorator(signal):
    called = False

    @receiver(signal)
    async def foo():
        nonlocal called
        called = True

    await signal.call()
    assert called


async def test_call_arguments(signal):
    received_args, received_kwargs = None, None

    @receiver(signal)
    async def foo(*args, **kwargs):
        nonlocal received_args, received_kwargs
        received_args, received_kwargs = args, kwargs

    await signal.call("foo", "bar", spam="spam")
    assert received_args == ("foo", "bar")
    assert received_kwargs == {"spam": "spam"}


async def multiple_receivers(signal):
    foo_called = False
    bar_called = False

    @receiver(signal)
    async def foo():
        nonlocal foo_called
        foo_called = True

    @receiver(signal)
    async def bar():
        nonlocal bar_called
        bar_called = True

    await signal.call()

    assert all((foo_called, bar_called))


async def test_add_remove_service_with_signals(entrypoint: aiomisc.Entrypoint):
    all_services = list()

    async def pre_start_hook(*, entrypoint, services):
        for service in services:
            all_services.append(service)

    pre_start_services = list()

    class SimpleService(aiomisc.Service):
        async def start(self) -> None:
            nonlocal pre_start_services
            pre_start_services.append(self)

    entrypoint.pre_start = entrypoint.pre_start.copy()
    entrypoint.post_start = entrypoint.post_start.copy()

    entrypoint.pre_start.connect(pre_start_hook)
    entrypoint.post_start.connect(pre_start_hook)

    entrypoint.pre_start.freeze()
    entrypoint.post_start.freeze()

    service = SimpleService()

    await entrypoint.start_services(service)

    assert len(all_services) == 2
    assert all_services == [service, service]

    assert len(pre_start_services) == 1
    assert service in pre_start_services



import asyncio
import struct
import uuid
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List

import pytest

import aiomisc
from aiomisc.io import AsyncFileIO
from tests import unix_only

pytestmark = pytest.mark.catch_loop_exceptions


compressors = pytest.mark.parametrize(
    "compression",
    list(aiomisc.io.Compression),
    ids=list(map(lambda c: c.name.lower(), aiomisc.io.Compression)),
)


async def test_simple(event_loop, tmp_path):
    tdir = Path(tmp_path)

    async with aiomisc.io.async_open(
        tdir / "test", "w+", loop=event_loop
    ) as afp:
        await afp.open()

        assert await afp.writable()
        assert await afp.readable()

        assert await afp.tell() == 0

        await afp.write("foo")
        assert await afp.tell() == 3

        await afp.seek(0)
        assert await afp.read() == "foo"

        await afp.write("\nbar\n")
        await afp.seek(0)

        assert await afp.readline() == "foo\n"
        assert await afp.readline() == "bar\n"

        await afp.flush()

        assert await afp.read() == ""

    with pytest.raises(ValueError):
        assert await afp.readable()

    with pytest.raises(ValueError):
        assert await afp.writable()


async def test_ordering(event_loop, tmp_path):
    tdir = Path(tmp_path)

    async with aiomisc.io.async_open(
        tdir / "file", "wb+", loop=event_loop
    ) as afp:
        await afp.seek(4)
        assert await afp.tell() == 4

        await afp.write(struct.pack("!I", 65535))
        assert await afp.tell() == 8

    assert afp.closed()

    async with aiomisc.io.async_open(
        tdir / "file", "rb+", loop=event_loop
    ) as afp:
        assert (await afp.read(4)) == b"\0\0\0\0"
        assert await afp.tell() == 4

        assert (await afp.read(4)) == struct.pack("!I", 65535)
        assert await afp.tell() == 8


async def test_async_for(event_loop, tmp_path):
    tdir = Path(tmp_path)
    afp: AsyncFileIO[str]

    async with aiomisc.io.async_open(
        tdir / "path", "w", loop=event_loop
    ) as afp:
        await afp.write("foo\nbar\nbaz\n")

    with open(tdir / "path") as fp:
        expected = []

        for line in fp:
            expected.append(line)

    assert expected

    result: list[str] = []
    async with aiomisc.io.async_open(
        tdir / "path", "r", loop=event_loop
    ) as afp:
        async for line in afp:
            result.append(line)

    assert result == expected


@unix_only
async def test_async_for_parallel(event_loop):
    tmp = NamedTemporaryFile(prefix="test_io")

    with tmp:
        async with aiomisc.io.async_open(
            tmp.name, "w+", loop=event_loop
        ) as afp:
            for _ in range(1024):
                await afp.write(f"{uuid.uuid4().hex}\n")

        with open(tmp.name) as fp:
            expected = set()

            for line in fp:
                expected.add(line)

        assert expected

        async with aiomisc.io.async_open(tmp.name, "r", loop=event_loop) as afp:
            result = set()

            async def reader():
                async for line in afp:
                    result.add(line)

            await asyncio.gather(reader(), reader(), reader(), reader())

    assert result == expected


@unix_only
async def test_object(event_loop):
    with NamedTemporaryFile(prefix="test_io") as tmp:
        afp1 = await aiomisc.io.async_open(tmp.name, "w+", loop=event_loop)
        afp2 = await aiomisc.io.async_open(tmp.name, "w+", loop=event_loop)

        async with afp1, afp2:
            assert afp1 == afp2
            assert hash(afp1) != hash(afp2)
            assert afp2 not in {afp1}
            assert afp1 in {afp1}

            with pytest.raises(TypeError):
                assert afp1 > afp2

            with pytest.raises(TypeError):
                assert afp1 < afp2

            for afp in (afp1, afp2):
                assert hash(afp) == hash(afp)
                assert isinstance(afp.fileno(), int)
                assert isinstance(afp.mode, str)
                assert isinstance(afp.name, str)
                assert isinstance(afp.errors, str)
                assert isinstance(afp.line_buffering, bool)
                assert afp.newlines is None


@compressors
async def test_compression(compression: aiomisc.io.Compression, tmp_path):
    fname = Path(tmp_path) / "test.file"

    async with aiomisc.io.async_open(
        fname, "w", compression=compression
    ) as afp:
        for i in range(1000):
            await afp.write(f"{i}\n")

    async with aiomisc.io.async_open(
        fname, "r", compression=compression
    ) as afp:
        for i in range(1000):
            assert await afp.readline() == f"{i}\n"



import asyncio
from collections.abc import Callable, Coroutine
from functools import wraps
from typing import Any, ParamSpec, TypeVar

T = TypeVar("T")
P = ParamSpec("P")
Number = int | float


def timeout(
    value: Number,
) -> Callable[
    [Callable[P, Coroutine[Any, Any, T]]], Callable[P, Coroutine[Any, Any, T]]
]:
    def decorator(
        func: Callable[P, Coroutine[Any, Any, T]],
    ) -> Callable[P, Coroutine[Any, Any, T]]:
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("Function is not a coroutine function")

        @wraps(func)
        async def wrap(*args: P.args, **kwargs: P.kwargs) -> T:
            return await asyncio.wait_for(func(*args, **kwargs), timeout=value)

        return wrap

    return decorator



import logging
from typing import Any

try:
    from logging_journald import JournaldLogHandler  # type: ignore

    def journald_formatter(**_: Any) -> logging.Handler:
        if not JournaldLogHandler.SOCKET_PATH.exists():
            raise FileNotFoundError("JournalD socket doesn't exists")

        handler = JournaldLogHandler()
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        return handler

except ImportError:

    def journald_formatter(**_: Any) -> logging.Handler:
        raise ImportError(
            'You must install "logging-journald" library for use it'
        )



import asyncio
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Awaitable, Callable, Coroutine
from contextlib import AbstractAsyncContextManager
from random import random
from typing import Any, Generic, NoReturn, TypeVar

from .compat import EventLoopMixin
from .utils import cancel_tasks

T = TypeVar("T", bound=Any)
Number = int | float

log = logging.getLogger(__name__)


class ContextManager(AbstractAsyncContextManager):
    __slots__ = "__aenter", "__aexit", "__instance"

    sentinel = object()

    def __init__(
        self,
        aenter: Callable[..., Awaitable[T]],
        aexit: Callable[..., Awaitable[T]],
    ):
        self.__aenter = aenter
        self.__aexit = aexit
        self.__instance = self.sentinel

    async def __aenter__(self) -> T:
        if self.__instance is not self.sentinel:
            raise RuntimeError("Reuse of context manager is not acceptable")

        self.__instance = await self.__aenter()
        return self.__instance

    async def __aexit__(
        self, exc_type: Any, exc_value: Any, traceback: Any
    ) -> Any:
        await self.__aexit(self.__instance)


class PoolBase(ABC, EventLoopMixin, Generic[T]):
    __slots__ = (
        "_create_lock",
        "_instances",
        "_recycle",
        "_recycle_bin",
        "_recycle_times",
        "_semaphore",
        "_tasks",
        "_len",
        "_used",
    ) + EventLoopMixin.__slots__

    _tasks: set[Any]
    _used: set[Any]
    _instances: asyncio.Queue
    _recycle_bin: asyncio.Queue

    def __init__(self, maxsize: int = 10, recycle: int | None = None):
        assert recycle is None or recycle > 0, (
            "recycle should be positive number or None"
        )

        self._instances = asyncio.Queue()
        self._recycle_bin = asyncio.Queue()

        self._semaphore = asyncio.Semaphore(maxsize)
        self._len = 0

        self._recycle = recycle

        self._tasks = set()
        self._used = set()

        self._create_lock = asyncio.Lock()
        self._recycle_times: defaultdict[float, Any] = defaultdict(
            self.loop.time
        )

        self.__create_task(self.__recycler())

    def __create_task(self, coro: Coroutine[Any, Any, Any]) -> asyncio.Task:
        task = self.loop.create_task(coro)
        self._tasks.add(task)
        task.add_done_callback(self._tasks.discard)
        return task

    async def __recycler(self) -> NoReturn:
        while True:
            instance = await self._recycle_bin.get()

            try:
                await self._destroy_instance(instance)
            except Exception:
                log.exception("Error when recycle instance %r", instance)
            finally:
                self._recycle_bin.task_done()

    @abstractmethod
    async def _create_instance(self) -> T:
        pass

    @abstractmethod
    async def _destroy_instance(self, instance: Any) -> None:
        pass

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    @abstractmethod
    async def _check_instance(self, instance: Any) -> bool:
        return True

    def __len__(self) -> int:
        return self._len

    def __recycle_instance(self, instance: Any) -> None:
        self._len -= 1
        self._semaphore.release()
        if instance in self._recycle_times:
            self._recycle_times.pop(instance)

        if instance in self._used:
            self._used.discard(instance)

        self._recycle_bin.put_nowait(instance)

    async def __create_new_instance(self) -> None:
        await self._semaphore.acquire()

        instance: Any = await self._create_instance()
        self._len += 1

        if self._recycle:
            deadline = self._recycle * (1 + random())
            self._recycle_times[instance] += deadline

        await self._instances.put(instance)

    async def __acquire(self) -> T:
        if not self._semaphore.locked():
            await self.__create_new_instance()

        instance = await self._instances.get()

        try:
            result = await self._check_instance(instance)
        except Exception:
            log.exception("Check instance %r failed", instance)
            self.__recycle_instance(instance)
        else:
            if not result:
                self.__recycle_instance(instance)
                return await self.__acquire()

        self._used.add(instance)
        return instance

    async def __release(self, instance: Any) -> None:
        self._used.discard(instance)

        if self._recycle and self._recycle_times[instance] < self.loop.time():
            self.__recycle_instance(instance)
            return

        self._instances.put_nowait(instance)

    def acquire(self) -> AbstractAsyncContextManager[T]:
        return ContextManager(self.__acquire, self.__release)

    async def close(self, timeout: Number | None = None) -> None:
        instances = list(self._used)
        self._used.clear()

        while self._instances.qsize():
            try:
                instances.append(self._instances.get_nowait())
            except asyncio.QueueEmpty:
                break

        async def log_exception(coro: Awaitable[Any]) -> None:
            try:
                await coro
            except Exception:
                log.exception("Exception when task execution")

        await asyncio.wait_for(
            asyncio.gather(
                *[
                    self.__create_task(
                        log_exception(self._destroy_instance(instance))
                    )
                    for instance in instances
                ],
                return_exceptions=True,
            ),
            timeout=timeout,
        )

        await cancel_tasks(self._tasks)



from .color import color_formatter
from .journald import journald_formatter
from .json import json_handler
from .rich import rich_formatter

__all__ = (
    "color_formatter",
    "journald_formatter",
    "json_handler",
    "rich_formatter",
)



import asyncio
import os
import signal
from time import sleep, time

import pytest

import aiomisc
from aiomisc.process_pool import ProcessPoolExecutor

POOL_SIZE = 4


@pytest.fixture
def pool():
    with ProcessPoolExecutor(POOL_SIZE) as pool:
        yield pool


@aiomisc.timeout(10)
async def test_simple(pool, event_loop, timer):
    current_time = await event_loop.run_in_executor(pool, time)
    assert current_time > 0

    with timer(1):
        await asyncio.wait_for(
            asyncio.gather(
                *[
                    event_loop.run_in_executor(pool, sleep, 1)
                    for _ in range(POOL_SIZE)
                ]
            ),
            timeout=2,
        )


@aiomisc.timeout(10)
async def test_exception(pool, event_loop):
    with pytest.raises(ZeroDivisionError):
        await event_loop.run_in_executor(pool, divmod, 1, 0)


def suicide():
    os.kill(os.getpid(), signal.SIGINT)


@aiomisc.timeout(10)
async def test_exit(pool, event_loop):
    await asyncio.wait_for(
        asyncio.gather(
            *[event_loop.run_in_executor(pool, suicide) for _ in range(4)],
            return_exceptions=True,
        ),
        timeout=2,
    )



import os
import platform
from multiprocessing import Queue
from pathlib import Path
from typing import Any

import pytest

import aiomisc
from aiomisc import threaded, timeout
from aiomisc.service import ProcessService, RespawningProcessService

pytestmark = pytest.mark.skipif(
    platform.system() == "Windows", reason="Temporary skip on windows"
)


def test_abstractmethod_exception():
    with pytest.raises(TypeError):
        ProcessService()  # type: ignore


class SampleProcessService(ProcessService):
    __required__ = ("path",)

    path: Path

    def in_process(self) -> Any:
        with open(self.path, "w") as fp:
            fp.write("Hello world\n")


def test_process_service(tmpdir):
    tmp_path = Path(tmpdir)
    test_file = tmp_path / "test.txt"
    svc = SampleProcessService(path=test_file)

    with aiomisc.entrypoint(svc) as loop:
        loop.run_until_complete(
            loop.run_in_executor(None, svc._process_stop_event.wait)
        )

    with open(test_file) as fp:
        assert fp.readline() == "Hello world\n"


class SimpleRespawningProcessService(RespawningProcessService):
    __required__ = ("queue",)

    queue: Queue

    def in_process(self) -> Any:
        self.queue.put(os.getpid())


def test_respawning_process_service():
    queue: Queue = Queue()
    svc = SimpleRespawningProcessService(queue=queue, process_poll_timeout=0.5)

    @timeout(5)
    async def go():
        pids = []

        @threaded
        def getter():
            return queue.get()

        for _ in range(2):
            pids.append(await getter())

        assert len(pids) == 2
        assert pids[0] != pids[1]

    with aiomisc.entrypoint(svc) as loop:
        loop.run_until_complete(go())



#!/usr/share/python3/doh/bin/python3
import logging

import aiohttp
import argclass
from yarl import URL

import aiomisc
from aiomisc.service.sdwatchdog import SDWatchdogService
from aiomisc.service.udp import UDPServer

log = logging.getLogger(__name__)


class Parser(argclass.Parser):
    address: str = "127.0.0.1"
    port: int = 25353
    log_level = argclass.LogLevel
    doh_url: URL = URL("https://1.1.1.1/dns-query")


class DNSUDPService(UDPServer):
    url: URL
    session: aiohttp.ClientSession

    async def handle_datagram(self, data: bytes, addr: tuple) -> None:
        request = self.session.post(
            self.url,
            data=data,
            headers={
                aiohttp.hdrs.ACCEPT: "application/dns-message",
                aiohttp.hdrs.CONTENT_TYPE: "application/dns-message",
            },
        )
        async with request as response:
            log.debug("Handling request from %r", addr)
            self.sendto(await response.read(), addr)

    async def start(self) -> None:
        await super().start()
        self.session = aiohttp.ClientSession()

    async def stop(self, exc: Exception = None) -> None:
        await self.session.close()
        await super().stop(exc)


if __name__ == "__main__":
    parser = Parser(auto_env_var_prefix="DOH_")
    parser.parse_args()

    services = [
        DNSUDPService(
            address=parser.address, port=parser.port, url=parser.doh_url
        ),
        SDWatchdogService(),
    ]

    with aiomisc.entrypoint(*services, log_level=parser.log_level) as loop:
        loop.run_forever()



import asyncio
import logging
import os
import socket
from collections.abc import Iterator
from importlib.metadata import Distribution, EntryPoint
from time import time_ns
from typing import Any, Concatenate, ParamSpec, Protocol, TypeAlias, final

from ._context_vars import EVENT_LOOP

log = logging.getLogger(__name__)


class EntrypointProtocol(Protocol):
    @property
    def name(self) -> str: ...

    def load(self) -> Any: ...


def entry_pont_iterator(entry_point: str) -> Iterator[EntrypointProtocol]:
    ep: EntryPoint
    for dist in Distribution.discover():
        for ep in dist.entry_points:
            if ep.group == entry_point:
                yield ep


class EventLoopMixin:
    __slots__ = ("_loop",)

    _loop: asyncio.AbstractEventLoop | None

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if not getattr(self, "_loop", None):
            self._loop = asyncio.get_running_loop()
        return self._loop  # type: ignore


event_loop_policy: asyncio.AbstractEventLoopPolicy
try:
    import uvloop

    if os.getenv("AIOMISC_USE_UVLOOP", "1").lower() in (
        "yes",
        "1",
        "enabled",
        "enable",
        "on",
        "true",
    ):
        event_loop_policy = uvloop.EventLoopPolicy()
    else:
        event_loop_policy = asyncio.DefaultEventLoopPolicy()
except ImportError:
    event_loop_policy = asyncio.DefaultEventLoopPolicy()


if hasattr(socket, "TCP_NODELAY"):

    def sock_set_nodelay(sock: socket.socket) -> None:
        if sock.proto != socket.IPPROTO_TCP:
            return None
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
else:

    def sock_set_nodelay(sock: socket.socket) -> None:
        return None


if hasattr(socket, "SO_REUSEPORT"):

    def sock_set_reuseport(sock: socket.socket, reuse_port: bool) -> None:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, reuse_port.real)
else:

    def sock_set_reuseport(sock: socket.socket, reuse_port: bool) -> None:
        log.debug(
            "SO_REUSEPORT is not implemented by underlying library. Skipping."
        )


# 16.x.x reverse compatibility
set_current_loop = EVENT_LOOP.set
get_current_loop = EVENT_LOOP.get

__all__ = (
    "Concatenate",
    "EntrypointProtocol",
    "EventLoopMixin",
    "ParamSpec",
    "Protocol",
    "TypeAlias",
    "entry_pont_iterator",
    "event_loop_policy",
    "final",
    "get_current_loop",
    "set_current_loop",
    "sock_set_nodelay",
    "sock_set_reuseport",
    "time_ns",
)



import asyncio
import time

import pytest

import aiomisc


async def test_kwargs(event_loop):
    mana = 0

    @aiomisc.asyncbackoff(
        attempt_timeout=0.5, deadline=0.5, pause=0, exceptions=(Exception,)
    )
    async def test():
        nonlocal mana

        if mana < 500:
            mana += 1
            await asyncio.sleep(5)
            raise ValueError("Not enough mana")

    t = time.time()
    with pytest.raises(asyncio.TimeoutError):
        await test()

    t2 = time.time() - t
    assert t2 > 0.4
    with pytest.raises(asyncio.TimeoutError):
        await test()

    t3 = time.time() - t
    assert t3 > 0.8

    assert mana < 3.8


async def test_simple(event_loop):
    mana = 0

    @aiomisc.asyncbackoff(0.10, 1, 0, Exception)
    async def test():
        nonlocal mana

        if mana < 5:
            mana += 1
            await asyncio.sleep(0.05)
            raise ValueError("Not enough mana")

    await test()

    assert mana == 5


async def test_simple_fail(event_loop):
    mana = 0

    @aiomisc.asyncbackoff(0.10, 0.5)
    async def test():
        nonlocal mana

        if mana < 500:
            mana += 1
            await asyncio.sleep(0.05)
            raise ValueError("Not enough mana")

    with pytest.raises(ValueError):
        await test()

    assert mana


async def test_too_long(event_loop):
    mana = 0

    @aiomisc.asyncbackoff(0.5, 0.5, 0, Exception)
    async def test():
        nonlocal mana

        if mana < 500:
            mana += 1
            await asyncio.sleep(5)
            raise ValueError("Not enough mana")

    with pytest.raises(asyncio.TimeoutError):
        await test()

    assert mana < 2


async def test_too_long_multiple_times(event_loop):
    mana = 0
    deadline = 0.5
    attempt_time = 0.06

    @aiomisc.asyncbackoff(attempt_time, deadline, 0, Exception)
    async def test():
        nonlocal mana

        if mana < 500:
            mana += 1
            await asyncio.sleep(5)
            raise ValueError("Not enough mana")

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(test(), timeout=2)

    assert mana < 11


async def test_exit(event_loop):
    mana = 0

    @aiomisc.asyncbackoff(0.05, 0)
    async def test():
        nonlocal mana

        if mana < 500:
            mana += 1
            await asyncio.sleep(5)
            raise ValueError("Not enough mana")

    with pytest.raises(asyncio.TimeoutError):
        await test()

    assert mana < 11


async def test_pause(event_loop):
    mana = 0
    condition = asyncio.Condition()

    @aiomisc.asyncbackoff(0.05, 1.2, 0.35)
    async def test():
        nonlocal mana

        async with condition:
            mana += 1
            condition.notify_all()

        await asyncio.sleep(0.2)
        raise ValueError("Not enough mana")

    task = event_loop.create_task(test())

    async with condition:
        await asyncio.wait_for(condition.wait_for(lambda: mana == 2), timeout=5)

    with pytest.raises(asyncio.TimeoutError):
        await task


async def test_no_waterline(event_loop):
    mana = 0
    condition = asyncio.Condition()

    @aiomisc.asyncbackoff(None, 1, 0, Exception)
    async def test():
        nonlocal mana

        async with condition:
            mana += 1
            await asyncio.sleep(0.1)
            condition.notify_all()
            raise ValueError("RETRY")

    task = event_loop.create_task(test())

    async with condition:
        await asyncio.wait_for(condition.wait_for(lambda: mana >= 5), timeout=2)

    with pytest.raises(ValueError, match="^RETRY$"):
        await task


@pytest.mark.parametrize("max_sleep", (0.5, 1))
async def test_no_deadline(event_loop, max_sleep):
    mana = 0
    condition = asyncio.Condition()

    @aiomisc.asyncbackoff(0.15, None, 0, Exception)
    async def test():
        nonlocal mana

        mana += 1
        await asyncio.sleep(max_sleep - (mana - 1) * 0.1)

        async with condition:
            condition.notify_all()

    task = event_loop.create_task(test())

    async with condition:
        await asyncio.wait_for(
            condition.wait_for(lambda: mana == max_sleep * 10),
            timeout=max_sleep * 10,
        )

    await asyncio.wait_for(task, timeout=max_sleep)


def test_values(event_loop):
    with pytest.raises(ValueError):
        aiomisc.asyncbackoff(-1, 1)

    with pytest.raises(ValueError):
        aiomisc.asyncbackoff(0, -1)

    with pytest.raises(ValueError):
        aiomisc.asyncbackoff(0, 0, -0.1)

    with pytest.raises(TypeError):
        aiomisc.asyncbackoff(0, 0)(lambda x: None)  # type: ignore


async def test_too_long_multiple(event_loop):
    mana = 0

    @aiomisc.asyncbackoff(0.5, 0.5, 0, Exception)
    async def test():
        nonlocal mana

        if mana < 500:
            mana += 1
            await asyncio.sleep(5)
            raise ValueError("Not enough mana")

    t = time.time()
    with pytest.raises(asyncio.TimeoutError):
        await test()

    t2 = time.time() - t
    assert t2 > 0.4
    with pytest.raises(asyncio.TimeoutError):
        await test()

    t3 = time.time() - t
    assert t3 > 0.8

    assert mana < 3.8


@pytest.mark.parametrize("max_tries", (1, 2, 5))
async def test_max_tries(event_loop, max_tries):
    mana = 0

    @aiomisc.asyncbackoff(0.5, 0.5, 0, Exception, max_tries=max_tries)
    async def test():
        nonlocal mana
        mana += 1
        raise ValueError("RETRY")

    with pytest.raises(ValueError, match="^RETRY$"):
        await test()

    assert mana == max_tries


@pytest.mark.parametrize("max_mana", (1, 2, 5))
async def test_giveup(event_loop, max_mana):
    mana = 0
    giveup_exception = None

    def giveup(exception):
        nonlocal giveup_exception
        giveup_exception = exception
        return mana >= max_mana

    @aiomisc.asyncbackoff(0.5, 0.5, 0, Exception, giveup=giveup)
    async def test():
        nonlocal mana
        mana += 1
        raise ValueError("RETRY")

    with pytest.raises(ValueError, match="^RETRY$") as e:
        await test()

    assert giveup_exception is e.value
    assert mana == max_mana


async def test_last_exception_is_last(event_loop):
    mana = 0

    class TestExc(Exception):
        pass

    @aiomisc.asyncbackoff(1, 2, 0)
    async def test():
        nonlocal mana

        if mana == 0:
            mana += 1
            await asyncio.sleep(1.5)  # causes TimeoutError on first try

        raise TestExc  # this is last exception actually

    with pytest.raises(TestExc):
        await test()


async def test_asyncbackoff_retry(event_loop):
    mana = 0

    @aiomisc.asyncbackoff(None, None, 0, Exception, max_tries=5)
    async def test():
        nonlocal mana
        mana += 1
        raise ValueError("RETRY")

    with pytest.raises(ValueError, match="^RETRY$"):
        await test()

    assert mana == 5


async def test_asyncretry(event_loop):
    mana = 0

    @aiomisc.asyncretry(5)
    async def test():
        nonlocal mana
        mana += 1
        raise ValueError("RETRY")

    with pytest.raises(ValueError, match="^RETRY$"):
        await test()

    assert mana == 5


async def test_asyncretry_class(event_loop):
    mana = 0

    class Foo:
        @aiomisc.asyncretry(5)
        async def test(self):
            nonlocal mana
            mana += 1
            raise ValueError("RETRY")

    foo = Foo()

    with pytest.raises(ValueError, match="^RETRY$"):
        await foo.test()

    assert mana == 5



import asyncio
import socket
import ssl
from collections.abc import Iterable, Mapping
from contextlib import suppress
from typing import Any

from aiohttp.web import Application, AppRunner, BaseRunner, SockSite

from aiomisc.service.tls import PathOrStr, SSLOptions

from ..utils import bind_socket
from .base import Service

try:
    from aiohttp.web_log import AccessLogger
except ImportError:  # pragma: nocover
    from aiohttp.helpers import AccessLogger  # type: ignore


RunnerKwargsType = Mapping[str, Any] | Iterable[tuple[str, Any]]


class AIOHTTPService(Service):
    __async_required__: tuple[str, ...] = ("start", "create_application")

    site: SockSite
    runner: BaseRunner
    handler_cancellation: bool = True

    def __init__(
        self,
        address: str | None = "localhost",
        port: int | None = None,
        sock: socket.socket | None = None,
        shutdown_timeout: int = 5,
        handler_cancellation: bool = handler_cancellation,
        runner_kwargs: RunnerKwargsType | None = None,
        **kwds: Any,
    ):
        if not sock:
            if not (address and port):
                raise RuntimeError(
                    "You should pass socket instance or "
                    '"address" and "port" couple'
                )

            self.socket = bind_socket(
                address=address, port=port, proto_name="http"
            )

        elif not isinstance(sock, socket.socket):
            raise ValueError("sock must be socket instance")
        else:
            self.socket = sock

        self.shutdown_timeout = shutdown_timeout

        self.runner_kwargs: dict[str, Any] = dict(runner_kwargs or {})
        self.runner_kwargs.setdefault("access_log_class", AccessLogger)
        self.runner_kwargs.setdefault(
            "access_log_format", AccessLogger.LOG_FORMAT
        )
        self.runner_kwargs.setdefault(
            "handler_cancellation", handler_cancellation
        )

        super().__init__(**kwds)

    async def create_application(self) -> Application:
        return Application()

    async def create_site(self) -> SockSite:
        if getattr(self, "runner", None) is None:
            raise RuntimeError("runner already created")

        return SockSite(self.runner, self.socket)

    async def make_runner(self, application: Application) -> AppRunner:
        return AppRunner(
            application,
            shutdown_timeout=self.shutdown_timeout,
            **self.runner_kwargs,
        )

    async def start(self) -> None:
        if hasattr(self, "runner"):
            raise RuntimeError("Can not start twice")

        application: Application = await self.create_application()
        self.runner = await self.make_runner(application)
        await self.runner.setup()

        self.site = await self.create_site()
        await self.site.start()

    async def stop(self, exception: Exception | None = None) -> None:
        try:
            if self.site:
                await self.site.stop()
        finally:
            # Ensure all handlers are finished
            await asyncio.sleep(0)
            if hasattr(self, "runner"):
                # Avoid AttributeError in case server already dereferenced
                with suppress(AttributeError):
                    await self.runner.cleanup()


class AIOHTTPSSLService(AIOHTTPService):
    def __init__(
        self,
        cert: PathOrStr,
        key: PathOrStr,
        ca: PathOrStr | None = None,
        address: str | None = None,
        port: int | None = None,
        verify: bool = True,
        sock: socket.socket | None = None,
        shutdown_timeout: int = 5,
        require_client_cert: bool = False,
        **kwds: Any,
    ):
        super().__init__(
            address=address,
            port=port,
            sock=sock,
            shutdown_timeout=shutdown_timeout,
            **kwds,
        )

        self.__ssl_options = SSLOptions(
            cert, key, ca, verify, require_client_cert, ssl.Purpose.CLIENT_AUTH
        )

    async def create_site(self) -> SockSite:
        assert self.runner and self.socket

        return SockSite(
            self.runner,
            self.socket,
            ssl_context=await self.loop.run_in_executor(
                None, self.__ssl_options.create_context
            ),
        )



from aiomisc.counters import Statistic, get_statistics


class StatSimple(Statistic):
    test_prop: int
    foo: int
    bar: float


class StatChild(StatSimple):
    baz: int


def test_get_statistics():
    assert list(get_statistics(StatSimple, StatChild)) == []

    simple = StatSimple()

    assert list(get_statistics(StatSimple, StatChild))

    simple.test_prop += 1

    for klass, name, metric, value in get_statistics(StatSimple, StatChild):
        if metric != "test_prop":
            continue

        assert value == 1

    # decref all
    del simple, klass, metric, value

    assert list(get_statistics(StatSimple, StatChild)) == []


def test_inheritance():
    stat = StatChild()

    assert stat.test_prop == 0
    assert stat.foo == 0
    assert stat.bar == 0.0
    assert stat.baz == 0
